const char *WebCfgDVCA_html=
   "<?xml version='1.0' ?>\n"
   "<!DOCTYPE html PUBLIC '-//W3C//DTD XHTML 1.0 Transitional//EN' 'http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd'>\n"
   "<html xmlns='http://www.w3.org/1999/xhtml'>\n"
   "<head>\n"
   "  <meta http-equiv='Content-Type' content='text/html; charset=##CHARSET##'/>\n"
   "  <meta http-equiv='Cache-Control' content='no-store, no-cache, must-revalidate'/>\n"
   "  <meta http-equiv='Cache-Control' content='post-check=0, pre-check=0'/>\n"
   "  <link rel='shortcut icon' href='/WebCfgD/ico' type='image' />\n"
   "  <title>##TITLE##</title>\n"
   "\n"
   "  <style type='text/css'>\n"
   "    body { background-color: #E6E6E6; font-family: Verdana,Arial,Helvetica,sans-serif; font-size: 12px; }\n"
   "\n"
   "    #gen-pnl-up { position: absolute; left: 0px; top: 0px; width: 800px; height: 30px; padding-top: 1px; padding-left: 5px; font-size: 15px; overflow: hidden; }\n"
   "    #gen-pnl-up div { width: 300%; }\n"
   "    #gen-pnl-up button { padding: 1px 3px; border: 1px solid #a9a9a9; float: left; margin-right: 5px; }\n"
   "    #gen-pnl-up div a img { padding-right: 5px; padding-top: 2px; }\n"
   "    #gen-pnl-up div > img { padding-right: 5px; height: 28px; }\n"
   "    /*#gen-pnl-up a { padding: 1px 3px; }*/\n"
   "    /*#gen-pnl-up button.active { border: 1px solid black; }*/\n"
   "    #gen-pnl-up img { height: 22px; border: 0; float: left; /*padding-right: 5px; vertical-align: middle;*/ }\n"
   "    #gen-pnl-up span { display: table-cell; height: 28px; border: 0; vertical-align: middle; }\n"
   "\n"
   "    #gen-pnl-left { position: absolute; left: 0px; top: 30px; width: 246px; height: 471px; background-color: #F7FFF6; border: 2px ridge #F7FFF6; overflow: auto; text-align: left; vertical-align: top; white-space: nowrap; color: black; }\n"
   "    #gen-pnl-left ul { list-style-image: none; list-style-position: outside; list-style-type: none; background-color: #F7FFF6; background-repeat: repeat-y; background-attachment: scroll; background-position: left top; margin: 0; white-space: nowrap; }\n"
   "    ul.tree { padding: 0; }\n"
   "    ul.tree ul { padding: 0 0 0 17px; }\n"
   "    ul.tree li.select { background-color: #BDEBED; }\n"
   "    ul.tree a { text-decoration:none; color: black; }\n"
   "    ul.tree a.pm { cursor: pointer; }\n"
   "    ul.tree img { vertical-align: middle; height: 16px; }\n"
   "    a.active img, button.active img { cursor: pointer; }\n"
   "    a.inactive img, button.inactive img { cursor: default; filter: grayscale(100%) opacity(30%); }\n"
   "\n"
   "    #gen-pnl-right { position: absolute; left: 250px; top: 30px; width: 546px; height: 471px; background-color: #F7F7F1; border: 2px outset #F7F7F1; overflow: auto; margin-left: 5px; padding-left: 1px; padding-right: 2px; }\n"
   "    #pgIco { padding-right: 10px; }\n"
   "    #gen-pnl-right table.page { padding-top: 5px; }\n"
   "    #gen-pnl-right td.title { color: black; font-family: Arial; font-size: 20px; font-style: italic; text-align: left; padding-bottom: 5px; }\n"
   "    #gen-pnl-right td.title img { vertical-align: middle; height: 32px; }\n"
   "    #gen-pnl-right td.tabs span { padding-left: 5px; padding-right: 5px; padding-bottom: 2px; padding-top: 2px; border: 1px solid gray; white-space: nowrap; background-color: #DEDED9; cursor: pointer; }\n"
   "    #gen-pnl-right td.tabs span.active { background-color: #F7F7F1; cursor: default; border-bottom: 0px none; border-top: 2px solid green; }\n"
   "    #gen-pnl-right td.content { border-top: 2px ridge grey; text-align: left; vertical-align: top; padding: 5px; }\n"
   "    #gen-pnl-right td.content div.elem { margin-bottom: 2px; overflow: visible; width: auto; white-space: nowrap; padding-left: 10px; text-indent: -10px; }\n"
   "    #gen-pnl-right td.content div.elem span { padding-right: 5px; }\n"
   "    #gen-pnl-right td.content div.elem span.label { white-space: nowrap; }\n"
   "    #gen-pnl-right td.content div.elem span.const { font-weight: bold; white-space: normal; }\n"
   "    #gen-pnl-right td.content div.elem textarea { width: 98%; font: 15px 'Open Sans', monospace; }\n"
   "    #gen-pnl-right td.content div.elem div.TA_Cntr + div { padding-left: 10px; }\n"
   "    #gen-pnl-right td.content div.elem div.TA_Cntr { width: 98%; text-indent: 0; }\n"
   "    #gen-pnl-right td.content div.elem div.TA_Cntr textarea { display: block; white-space: pre; position: absolute; margin: 0; border-radius: 0; border: 2px solid #685972; padding: 2px; background-color: transparent; resize: none; caret-color: black; }\n"
   "    #gen-pnl-right td.content div.elem div.TA_Cntr div.TA_BD { display: block; position: absolute; overflow: auto; background-color: #fff; border: 2px solid #685972; padding: 2px; }\n"
   "    #gen-pnl-right td.content div.elem div.TA_Cntr div.TA_HGLs { white-space: pre; word-wrap: break-word; /*color: transparent;*/ font: 15px 'Open Sans', monospace; }\n"
   "    #gen-pnl-right td.content div.elem div.TA_Cntr div.TA_HGLs span { padding: 0; }\n"
   "    #gen-pnl-right td.content fieldset.elem { margin : 0px; margin-bottom: 2px; padding: 3px; }\n"
   "    #gen-pnl-right td.content div.table { width: 800px; border: 1px solid gray; height: 200px; background-color: white; overflow: auto; padding: 2px; }\n"
   "    #gen-pnl-right td.content table.elem { border: 1px solid black; border-collapse: collapse; empty-cells: show; text-indent: 0px; }\n"
   "    #gen-pnl-right td.content table.elem th { border: 1px solid black; background-color: #E6E6E6; text-align: center; white-space: pre-line; }\n"
   "    #gen-pnl-right td.content table.elem td { border: 1px solid black; white-space: pre-line; /*word-break: break-all;*/}\n"
   "    #gen-pnl-right td.content table.elem td.edit { white-space: nowrap; }\n"
   "    #gen-pnl-right td.content table.elem td input, #gen-pnl-right td.content table.elem td select, #gen-pnl-right td.content table.elem td textarea\n"
   "    { vertical-align: middle; border: 1px solid gray; width: 100%; box-sizing: border-box; }\n"
   "    #gen-pnl-right td.content table.elem td img { vertical-align: middle; height: 16px; cursor: pointer; }\n"
   "    #gen-pnl-right td.content table.elem tr.select td { background-color: #00C8FF; }\n"
   "    #gen-pnl-right td.content .list { width: 95%; }\n"
   "    #gen-pnl-right td.content .list option { white-space: normal; }\n"
   "    #gen-pnl-right td.content .picture { border: 1px solid blue; vertical-align: top; }\n"
   "    #gen-pnl-right td.content div.elem .line input[size=\"30\"] { width: 50%; }\n"
   "    #gen-pnl-right td.content div.elem .line { padding-right: 3px; }\n"
   "    #gen-pnl-right td.content div.elem .line input,select { vertical-align: middle; border: 1px solid gray; }\n"
   "    #gen-pnl-right td.content div.elem .line img { vertical-align: middle; height: 18px; cursor: pointer; }\n"
   "    #gen-pnl-right td.content div.elem .number input { /*width: 20%;*/ text-align: right; }\n"
   "\n"
   "    #gen-pnl-down { position: absolute; left: 0px; top: 500px; width: 800px; height: 25px; font-size: 15px; overflow: hidden; }\n"
   "    #gen-pnl-splt { position: absolute; left: 250px; top: 25px; width: 2px; height: 550px; border: 0; cursor: col-resize; margin-left: 2px; border-left: 2px ridge #E6E6E6; }\n"
   "    #st_user { border-left: 1px solid black; padding-left: 2px; white-space: nowrap; }\n"
   "\n"
   "    #gen-pnl-dlg > table { position: absolute; box-shadow: 10px 10px 5px rgba(68,68,68,0.6); z-index: 9999; background-color: #8FA9FF; border: 1px solid darkblue; }\n"
   "    #gen-pnl-dlg #wDlgHeader { padding-left: 5px; overflow: hidden; white-space: nowrap; background: linear-gradient(#8FA9FF, #6F89DF, #8FA9FF); }\n"
   "    #gen-pnl-dlg > div { position: absolute; left: 0px; top: 0px; width: 100%; height: 100%; opacity: 0.1; background-color: #8F8F8F; }\n"
   "    #gen-pnl-dlg .dlg { width: 100%; border: 2px ridge #FF9253; background-color: #E6E6E6; padding: 5px; text-align: left; vertical-align: top; }\n"
   "\n"
   "    #popupmenu { position: absolute; border: 0; width: 150px; height: 150px; overflow: auto; z-index: 10; }\n"
   "    #popupmenu select { background-color: #E6E6E6; border: 1px solid black; padding: 1px; width: inherit; height: inherit; }\n"
   "    #combomenu { position: absolute; border: 1px solid black; width: 100%; height: 120px; overflow: auto; background-color: white; padding: 1px; z-index: 10; }\n"
   "    #combomenu select { background-color: white; border: 0; width: 100%; height: 120px; }\n"
   "  </style>\n"
   "</head>\n"
   "<body alink='#33ccff' link='#3366ff' text='#000000' vlink='#339999'>\n"
   "<div id='gen-pnl-up'><div>\n"
   "  <button id='actLoad' class='inactive' title='###Load###'><img src='/WebCfgD/img_load' /></button>\n"
   "  <button id='actLoad_f' class='active' title='###Load forcibly###'><img src='/WebCfgD/img_load_f' /></button>\n"
   "  <button id='actSave' class='inactive' title='###Save###'><img src='/WebCfgD/img_save' /></button>\n"
   "  <button id='actSave_f' class='active' title='###Save forcibly###'><img src='/WebCfgD/img_save_f' /></button>\n"
   "  <img src='/WebCfgD/img_line' />\n"
   "  <button id='actUp' class='inactive' title='###Up###'><img src='/WebCfgD/img_up' /></button>\n"
   "  <button id='actPrevious' class='inactive' title='###Back###'><img src='/WebCfgD/img_previous' /></button>\n"
   "  <button id='actNext' class='inactive' title='###Forward###'><img src='/WebCfgD/img_next' /></button>\n"
   "  <img src='/WebCfgD/img_line'/>\n"
   "  <button id='actAddIt' class='inactive' title='###Add item###'><img src='/WebCfgD/img_it_add' /></button>\n"
   "  <button id='actDelIt' class='inactive' title='###Delete item###'><img src='/WebCfgD/img_it_del' /></button>\n"
   "  <img src='/WebCfgD/img_line'/>\n"
   "  <button id='actCopy' class='inactive' title='###Copy item###'><img src='/WebCfgD/img_editcopy' /></button>\n"
   "  <button id='actCut' class='inactive' title='###Cut item###'><img src='/WebCfgD/img_editcut' /></button>\n"
   "  <button id='actPaste' class='inactive' title='###Paste item###'><img src='/WebCfgD/img_editpaste' /></button>\n"
   "  <img src='/WebCfgD/img_line' />\n"
   "  <button id='actUpdate' class='active' title='###Refresh the item and tree###'><img src='/WebCfgD/img_reload' /></button>\n"
   "  <button id='actStart' class='active' title='###Start the cycled refreshing###'><img src='/WebCfgD/img_start' /></button>\n"
   "  <button id='actStop' class='inactive' title='###Stop the cycled refreshing###'><img src='/WebCfgD/img_stop' /></button>\n"
   "  <img src='/WebCfgD/img_line' />\n"
   "  <a id='actManual' class='inactive' title='###Manual###' target='_blank'><img src='/WebCfgD/img_manual' /></a>\n"
   "  <a id='actAbout' class='active' href='/WebCfgD/about' title='###About###'><img src='/WebCfgD/img_help' /></a>\n"
   "  <img src='/WebCfgD/img_line' />\n"
   "  <span id='selPath' />\n"
   "</div></div>\n"
   "<div id='gen-pnl-left'>\n"
   "  <div id='treeBlk'><ul id='treeRoot' class='tree'/></div>\n"
   "</div>\n"
   "<div id='gen-pnl-right'>\n"
   "  <table class='page' border='0' cellspacing='0' width='100%'>\n"
   "    <tr><td class='title'><img id='pgIco' alt='' /><span id='pgTitle' /></td></tr>\n"
   "    <tr><td id='pgTabs' class='tabs'></td></tr>\n"
   "    <tr><td class='content'><div id='pgCont'></div></td></tr>\n"
   "  </table>\n"
   "</div>\n"
   "<div id='gen-pnl-down'>\n"
   "<table width='100%'><TR><td id='status' width=\"100%\"/><td id='st_user'><a href=\"/login/WebCfgD/\">##USER##</a></td></TR></table>\n"
   "</div>\n"
   "<div id='gen-pnl-splt' title=\"###Drag to resize the Menu###\" />\n"
   "<script src='/WebCfgD/script.js' type='text/javascript'></script>\n"
   "</body>\n"
   "</html>\n";
